/*  Copyright (C) 2015-2020 Michele Santullo
 *
 *  This file is part of DuckScraper.
 *
 *  DuckScraper is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  DuckScraper is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with DuckScraper.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "htmlretrieve.hpp"
#include <ciso646>
#include <tidy.h>
#include <tidybuffio.h>
#include <sstream>
#include <curl_easy.h>
#include <cstring>
#include <stack>
#include <algorithm>
#include <memory>
#include <cassert>
#include <utility>
#include <cctype>
#include <iterator>

namespace duck {
	namespace {
		std::string make_lowercase (std::string_view in) {
			std::string out;
			std::transform(in.begin(), in.end(), std::back_inserter(out), [](unsigned char c){return std::tolower(c);});
			return out;
		}

		TidyEncodingOptions charset_to_enum (std::string_view name) {
			const std::string lower_name = make_lowercase(name);
			if (lower_name == "ascii")
				return TidyEncAscii;
			//else if (lower_name == "???")
			//	return TidyEncLatin0;
			//else if (lower_name == "???")
			//	return TidyEncLatin1;
			else if (lower_name == "utf-8")
				return TidyEncUtf8;
#ifndef NO_NATIVE_ISO2022_SUPPORT
			else if (lower_name == "iso-2022-cn")
				return TidyEncIso2022;
#endif
			else if (lower_name == "mac")
				return TidyEncMac;
			else if (lower_name == "windows-1252")
				return TidyEncWin1252;
			else if (lower_name == "ibm858")
				return TidyEncIbm858;
			else if (lower_name == "utf-16le")
				return TidyEncUtf16le;
			else if (lower_name == "utf-16be")
				return TidyEncUtf16be;
			else if (lower_name == "utf-16")
				return TidyEncUtf16;
			else if (lower_name == "big-5")
				return TidyEncBig5;
			else if (lower_name == "shift-jis" or lower_name == "shift_jis")
				return TidyEncShiftjis;
			throw std::runtime_error("Failed to recognise \"" + std::string(name) + "\" as a valid source charset");
		}

		bool isHttps (const std::string_view& parUrl) {
			const char protocol[] = "https://";
			const size_t protocolLen = sizeof(protocol) / sizeof(protocol[0]) - 1;
			if (parUrl.size() < protocolLen)
				return false;

			return std::equal(protocol, protocol + protocolLen, parUrl.begin());
		}
	} //unnamed namespace

	std::string clean_html (std::string&& html, OptString src_charset) {

		// Initialize a Tidy document
		TidyDoc tidyDoc = tidyCreate();
		TidyBuffer tidyOutputBuffer = {nullptr, nullptr, 0, 0, 0};

		// Configure Tidy
		// The flags tell Tidy to output XML and disable showing warnings
		bool configSuccess = tidyOptSetBool(tidyDoc, TidyXmlOut, yes)
			&& tidyOptSetBool(tidyDoc, TidyQuiet, yes)
			&& tidyOptSetBool(tidyDoc, TidyNumEntities, yes)
			&& tidyOptSetBool(tidyDoc, TidyShowWarnings, no)
			&& tidyOptSetBool(tidyDoc, TidyEscapeScripts, yes)
			&& tidyOptSetInt(tidyDoc, TidyNewline, TidyLF)
			&& tidyOptSetBool(tidyDoc,TidyFixUri, yes)
			&& tidyOptSetBool(tidyDoc, TidyHideComments, yes)
			&& tidyOptSetBool(tidyDoc, TidyPPrintTabs, yes)
			&& tidyOptSetBool(tidyDoc, TidyLiteralAttribs, no)
			&& tidyOptSetBool(tidyDoc, TidyPunctWrap, no)
			&& tidyOptSetInt(tidyDoc, TidyOutCharEncoding, TidyEncUtf8)
			&& tidyOptSetBool(tidyDoc, TidyMetaCharset, yes)
			&& tidyOptSetValue(tidyDoc, TidyDoctype, "omit")
			&& tidyOptSetBool(tidyDoc, TidyDropPropAttrs, yes);

		if (src_charset)
			tidyOptSetInt(tidyDoc, TidyInCharEncoding, charset_to_enum(*src_charset));

		int tidyResponseCode = -1;

		// Parse input
		if (configSuccess) {
			tidyResponseCode = tidyParseString(tidyDoc, html.c_str());
		}

		// Process HTML
		if (tidyResponseCode >= 0)
			tidyResponseCode = tidyCleanAndRepair(tidyDoc);

		if (tidyResponseCode >= 0)
			tidyResponseCode = tidyRunDiagnostics(tidyDoc);
		if (tidyResponseCode > 1)
			tidyResponseCode = (tidyOptSetBool(tidyDoc, TidyForceOutput, yes) ? tidyResponseCode : -1);

		// Output the HTML to our buffer
		if (tidyResponseCode >= 0)
			tidyResponseCode = tidySaveBuffer(tidyDoc, &tidyOutputBuffer);

		// Any errors from Tidy?
		if (tidyResponseCode < 0)
			throw ("Tidy encountered an error while parsing an HTML response. Tidy response code: " + tidyResponseCode);

		// Grab the result from the buffer and then free Tidy's memory
		std::string tidyResult = (char*)tidyOutputBuffer.bp;
		tidyBufFree(&tidyOutputBuffer);
		tidyRelease(tidyDoc);

		return tidyResult;
	}


	std::string fetch_html (const std::string_view& parSource, std::string parUserAgent, bool parSslVerifyPeer, bool parSslVerifyHost) {
		using curl::curl_easy;
		using curl::curl_pair;
		using curl::curl_ios;

		std::ostringstream oss;
		curl_ios<std::ostream> wr(oss);
		curl_easy easy(wr);
		easy.add(curl_pair<CURLoption, std::string>(CURLOPT_URL, std::string(parSource)));
		if (isHttps(parSource)) {
			easy.add(curl_pair<CURLoption, bool>(CURLOPT_SSL_VERIFYPEER, parSslVerifyPeer));
			easy.add(curl_pair<CURLoption, bool>(CURLOPT_SSL_VERIFYHOST, parSslVerifyHost));
		}
		easy.add(curl_pair<CURLoption, std::string>(CURLOPT_USERAGENT, parUserAgent));
		easy.add(curl_pair<CURLoption, long>(CURLOPT_FOLLOWLOCATION, 1L));

		easy.add(curl_pair<CURLoption, std::string>(CURLOPT_ACCEPT_ENCODING, "gzip"));
		easy.add(curl_pair<CURLoption, long>(CURLOPT_HTTP_CONTENT_DECODING, 1L));

		//try {
			easy.perform();
		//}
		//catch (curl_error& err) {
			//std::stack<std::pair<std::string, std::string> > errors = err.what();
			//err.print_traceback();
			//return 1;
		//}

		//return FetchedHtml(oss.str(), easy.get_info<CURLINFO_CONTENT_TYPE>().get());
		return oss.str();
	}
} //namespace duck
