/*  Copyright (C) 2015 Michele Santullo
 *
 *  This file is part of DuckScraper.
 *
 *  DuckScraper is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  DuckScraper is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with DuckScraper.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "iconv_wrapper.hpp"
#include <iconv.h>
#include <cassert>
#include <algorithm>

namespace duck {
	namespace {
		class IconvDeleter {
		public:
			typedef iconv_t pointer;

			void operator() (pointer& resource) {
				if (resource and reinterpret_cast<iconv_t>(-1) != resource) {
					iconv_close(resource);
					resource = static_cast<iconv_t>(0);
				}
			}
		};
	} //unnamed namespace

	typedef std::unique_ptr<iconv_t, IconvDeleter> UniqueIconv;

	IconvBadSequence::IconvBadSequence (const std::string& message) :
		std::domain_error(message)
	{
	}

	IconvOpenFailure::IconvOpenFailure (const std::string& message) :
		std::logic_error(message)
	{
	}

	struct IconvWrapper::LocalData {
		UniqueIconv context;
	};

	IconvWrapper::IconvWrapper(const std::string& from, std::string to, Mode mode) :
		m_local(std::make_unique<LocalData>())
	{
		switch (mode) {
		case ModeIgnore:
			to += "//IGNORE";
			break;
		case ModeTransliterate:
			to += "//TRANSLIT";
			break;
		case ModeDefault:
		default:
			break;
		}

		m_local->context.reset(iconv_open(to.c_str(), from.c_str()));
		if (reinterpret_cast<iconv_t>(-1) == m_local->context.get()) {
			auto msg = std::string("Failed to create an iconv context for \"") +
				from + "\" to \"" + to + "\" conversion (error code " +
				std::to_string(errno);
			if (EINVAL == errno)
				msg += " EINVAL";
			msg += ")";

			throw IconvOpenFailure(msg);
		}
	}

	IconvWrapper::~IconvWrapper() noexcept = default;

	void IconvWrapper::conv (
		const char* buff,
		std::size_t len,
		PtrGetterFunc get_ptr,
		SizeGetterFunc get_size,
		ReallocFunc realloc,
		std::size_t grow_hint
	) {
		assert(buff);
		assert(len);

		const constexpr std::size_t def_inc = 16;
		const constexpr std::size_t iconv_err = static_cast<std::size_t>(-1);

		std::size_t nchars;
		std::size_t inbytesleft = len;
		char* inbuff = const_cast<char*>(buff);
		std::ptrdiff_t out_offset = 0;
		std::size_t grow_factor = grow_hint;
		std::size_t outbytesleft;

		do {
			realloc(std::max(inbytesleft * grow_factor, def_inc) + get_size());
			assert(get_size() > static_cast<std::size_t>(out_offset));
			outbytesleft = get_size() - out_offset;
			char* outbuff = get_ptr() + out_offset;

			const auto old_inbytesleft = inbytesleft;
			const auto old_outbytesleft = outbytesleft;

			nchars = ::iconv(
				m_local->context.get(),
				&inbuff,
				&inbytesleft,
				&outbuff,
				&outbytesleft
			);
			if (iconv_err == nchars) {
				const auto pos_str = std::to_string(len - inbytesleft);
				switch (errno) {
				case EILSEQ:
					throw IconvBadSequence("Invalid input multibyte sequence at byte " + pos_str);
				case EINVAL:
					throw IconvBadSequence("Incomplete input multibyte sequence at byte " + pos_str);
				}
			}

			out_offset = std::distance(get_ptr(), outbuff);
			assert(out_offset >= 0);

			const auto in_diff = old_inbytesleft - inbytesleft;
			const auto out_diff = old_outbytesleft - outbytesleft;
			grow_factor = std::max<std::size_t>(1, out_diff / in_diff);
		} while (iconv_err == nchars and E2BIG == errno);

		assert(outbytesleft < get_size());
		realloc(get_size() - outbytesleft);
	}
} //namespace duck
