/*  Copyright (C) 2015 Michele Santullo
 *
 *  This file is part of DuckScraper.
 *
 *  DuckScraper is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  DuckScraper is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with DuckScraper.  If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include <memory>
#include <string>
#include <string_view>
#include <functional>
#include <stdexcept>

namespace duck {
	class IconvBadSequence : public std::domain_error {
	public:
		explicit IconvBadSequence(const std::string& message);
	};
	class IconvOpenFailure : public std::logic_error {
	public:
		explicit IconvOpenFailure(const std::string& message);
	};

	class IconvWrapper {
		typedef std::function<char*()> PtrGetterFunc;
		typedef std::function<void(std::size_t)> ReallocFunc;
		typedef std::function<std::size_t()> SizeGetterFunc;
	public:
		enum Mode {
			ModeTransliterate, ModeIgnore, ModeDefault
		};

		IconvWrapper (const std::string& from, std::string to, Mode mode=ModeDefault);
		~IconvWrapper() noexcept;

		template <typename CIn, typename COut>
		std::basic_string<COut> conv (std::basic_string_view<CIn> text);

		std::string conv_char (std::string_view text) {return conv<char, char>(text);}

	private:
		struct LocalData;

		void conv (
			const char* buff,
			std::size_t len,
			PtrGetterFunc get_ptr,
			SizeGetterFunc get_size,
			ReallocFunc realloc,
			std::size_t grow_hint
		);

		std::unique_ptr<LocalData> m_local;
	};

	template <typename CIn, typename COut>
	std::basic_string<COut> IconvWrapper::conv (std::basic_string_view<CIn> text) {
		typedef std::basic_string<COut> string;

		if (text.empty())
			return {};

		string retval;
		this->conv(
			reinterpret_cast<const char*>(text.data()),
			text.size() * sizeof(CIn),
			[&retval](){return reinterpret_cast<char*>(retval.data());},
			[&retval]()->std::size_t {return retval.size() * sizeof(COut);},
			[&retval](std::size_t sz){retval.resize((sz + sizeof(COut) - 1) / sizeof(COut));},
			sizeof(COut) / sizeof(CIn)
		);
		return retval;
	}
} //namespace duck
