/*  Copyright (C) 2015 Michele Santullo
 *
 *  This file is part of DuckScraper.
 *
 *  DuckScraper is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  DuckScraper is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with DuckScraper.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "commandline.hpp"
#include "xpath.hpp"
#include "scraplang.hpp"
#include "html_pool.hpp"
#include "read_all.hpp"
#include "safe_stack_object.hpp"
#include <iostream>
#include <string>
#include <fstream>
#include <utility>
#include <ciso646>
#include <memory>
#include <stdexcept>

namespace {
	void dump_string ( const std::string& parPathDest, const std::string& parData );
	void load_from_commandline ( const boost::program_options::variables_map& parVarMap, duck::XPathPtr xpath );
	void load_from_model ( const boost::program_options::variables_map& parVarMap, duck::XPathPtr xpath );
} //unnamed namespace

int main (int argc, char* argv[]) {
	using boost::program_options::variables_map;

	variables_map vm;
	try {
		if (duck::parse_commandline(argc, argv, vm)) {
			return 0;
		}
	}
	catch (const std::invalid_argument& err) {
		std::cerr << err.what() << "\nUse --help for help" << std::endl;
		return 2;
	}

	try {
		curry::SafeStackObject<duck::XPath> query;
		if (vm.count("model"))
			load_from_model(vm, query);
		else
			load_from_commandline(vm, query);
	}
	catch (const duck::ParseError& err) {
		std::cerr << err.what() << std::endl;
		return 1;
	}
	catch (const std::runtime_error& err) {
		std::cerr << err.what() << std::endl;
		return 1;
	}
	return 0;
}

namespace {
	void dump_string (const std::string& parPathDest, const std::string& parData) {
		std::unique_ptr<std::ofstream> ofs;
		const bool use_stdout = ("-" == parPathDest);
		if (not use_stdout) {
			ofs.reset(new std::ofstream(parPathDest));
		}
		std::ostream* const os = (use_stdout ? &std::cout : ofs.get());
		*os << parData;
	}

	void load_from_commandline (const boost::program_options::variables_map& parVarMap, duck::XPathPtr xpath) {
		using std::string;

		const auto& vm = parVarMap;
		const auto url = vm["input-url"].as<string>();

		duck::HtmlPool html_pool(
			string(parVarMap["agent"].as<string>()),
			string(parVarMap["from-code"].as<string>())
		);
		const auto in_html_id = html_pool.GetOrAdd(url);
		string html = *html_pool.GetByID(in_html_id);
		if (vm.count("dump")) {
			dump_string(vm["dump"].as<string>(), html);
		}

		const string xpath_str = parVarMap["xpath"].as<string>();

#if !defined(NDEBUG)
		std::cout << " -- XPath direct mode --\n";
		std::cout << "URL  : " << parVarMap["input-url"].as<string>() << "\n";
		std::cout << "XPath: " << xpath_str << std::endl;
		std::cout << "Agent: " << parVarMap["agent"].as<string>() << std::endl;
#endif

		std::vector<string> queries;
		queries.reserve(1);
		queries.push_back(std::move(xpath_str));
		auto results = xpath->run_query(html, queries, parVarMap["namespace"].as<string>());
		for (const auto& lst : results[0]) {
			std::cout << lst.first << ": " << lst.second << '\n';
		}
	}

	void load_from_model (const boost::program_options::variables_map& parVarMap, duck::XPathPtr xpath) {
		using std::string;

#if !defined(NDEBUG)
		std::cout << " -- XPath model mode --\n";
		if (parVarMap.count("input-url"))
			std::cout << "URL  : " << parVarMap["input-url"].as<string>() << "\n";
		std::cout << "Model: " << parVarMap["model"].as<string>() << std::endl;
		std::cout << "Agent: " << parVarMap["agent"].as<string>() << std::endl;
#endif
		const string script = duck::read_all(parVarMap["model"].as<string>());
		auto ast = duck::sl::parse(script);

		duck::HtmlPool html_pool(
			string(parVarMap["agent"].as<string>()),
			string(parVarMap["from-code"].as<string>())
		);
		duck::sl::apply(ast, duck::sl::HtmlPoolBaseSP(&html_pool), xpath, string(parVarMap["namespace"].as<string>()));
		//auto list = duck::get_xpath_definitions(*ast);

		//std::vector<string> expressions;
		//expressions.reserve(list.size());
		//for (duck::element_def& elem : list) {
		//	expressions.push_back(std::move(elem.xpath));
		//}
		//auto results = duck::xpath_query(parXML, expressions);
		//duck::print_results(std::cout, *ast, list, results);
		//for (const auto& list : results) {
		//	std::cout << "------\n";
		//	for (const auto& result : list) {
		//		std::cout << result.first << ": " << result.second << '\n';
		//	}
		//}
	}
} //unnamed namespace
