/*  Copyright (C) 2015 Michele Santullo
 *
 *  This file is part of DuckScraper.
 *
 *  DuckScraper is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  DuckScraper is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with DuckScraper.  If not, see <http://www.gnu.org/licenses/>.
 */

#define APPLY_VERBOSE

#include "apply.hpp"
#include "mstch/mstch.hpp"
#include "html_pool_base.hpp"
#include "scrap_node.hpp"
#include "xpath_runner.hpp"
#if defined(APPLY_VERBOSE)
#	include "stream_scrap_node.hpp"
#endif
#include <map>
#include <boost/variant/apply_visitor.hpp>
#include <string_view>
#include <list>
#include <functional>
#include <algorithm>
#include <cassert>

namespace duck { namespace sl {
#if defined(APPLY_VERBOSE)
#endif

	namespace {
		struct EntryNode;

		struct MustacheEntry {
			std::string text;
			mstch::map context;
		};

		using EntryNodeList = std::vector<std::pair<
			const SourceInfo*,
			EntryNode
		>>;
		using MustacheEntryMap = std::map<std::string, MustacheEntry>;
		typedef std::function<void(std::size_t)> FixLengthCommand;

		struct EntryNode {
			explicit EntryNode (const std::string_view& parName) :
				name(parName)
			{
			}
			EntryNode (EntryNode&&) = default;
			EntryNode (const EntryNode&) = default;
			EntryNode& operator= (EntryNode&&) = default;
			EntryNode& operator= (const EntryNode&) = default;

			std::string_view name;
			std::vector<EntryNode> structs;
			std::vector<const XPathElement*> xpaths;
		};

		struct ApplyEntry {
			ApplyEntry (const ApplyEntry&) = default;
			ApplyEntry (const SourceInfo* parAppTo, std::string_view parMstchName) :
				apply_to(parAppTo),
				content(""),
				mustache_name(parMstchName)
			{
				assert(apply_to);
				assert(not apply_to->value.empty());
			}
			ApplyEntry (ApplyEntry&&) = default;
			ApplyEntry& operator=(ApplyEntry&&) = default;

			const SourceInfo* apply_to;
			EntryNode content;
			std::string_view mustache_name;
		};

		class StructItemExtractor : public boost::static_visitor<> {
		public:
			StructItemExtractor() = delete;
			explicit StructItemExtractor (EntryNode& parRoot) :
				m_root(parRoot)
			{
			}

			void operator() (const XPathElement& parVal) {
				m_root.xpaths.push_back(&parVal);
			}

			void operator() (const StructBlock& parVal) {
				m_root.structs.emplace_back(parVal.name);
				StructItemExtractor visitor(m_root.structs.back());
				for (auto& itm : parVal.xpaths) {
					boost::apply_visitor(visitor, itm);
				}
			}

		private:
			EntryNode& m_root;
		};

		class FillWithClonesVisitor : public boost::static_visitor<mstch::node&&> {
		public:
			explicit FillWithClonesVisitor (std::size_t exp_size) :
				m_expected_size(exp_size)
			{ }
			virtual mstch::node&& operator()(mstch::array&& parOut) {
				if (parOut.empty()) {
					parOut.resize(m_expected_size);
				}
				else {
					std::fill_n(
						std::back_inserter(parOut),
						std::max(m_expected_size, parOut.size()) - parOut.size(),
						parOut.back()
					);
				}
				m_retval = std::move(parOut);
				return std::move(m_retval);
			}

			template <typename T>
			mstch::node&& operator()(T&& parOut) {
				mstch::array retval;
				retval.reserve(m_expected_size);
				retval.push_back(std::move(parOut));
				return (*this)(std::move(retval));
			}

		protected:
			mstch::node m_retval;
			std::size_t m_expected_size;
		};

		class FillWithStringsVisitor : public FillWithClonesVisitor {
		public:
			FillWithStringsVisitor (const std::string& parVal, std::size_t exp_size) :
				FillWithClonesVisitor(exp_size),
				m_value(parVal)
			{ }

			mstch::node&& operator()(mstch::array&& parOut) override {
				std::fill_n(
					std::back_inserter(parOut),
					std::max(m_expected_size, parOut.size()) - parOut.size(),
					m_value.get()
				);
				m_retval = std::move(parOut);
				return std::move(m_retval);
			}

			using FillWithClonesVisitor::operator();

		private:
			std::reference_wrapper<const std::string> m_value;
		};

		void fill_with_defaults (mstch::map& parMap, std::size_t parTotal, const std::string& parKey, const std::string& parDefault) {
			FillWithStringsVisitor visitor(parDefault, parTotal);
			parMap[parKey] = boost::apply_visitor(visitor, std::move(parMap[parKey]));
		}

		void fill_with_last_item_clones (mstch::map& parMap, std::size_t parTotal, const std::string& parKey) {
			FillWithClonesVisitor visitor(parTotal);
			parMap[parKey] = boost::apply_visitor(visitor, std::move(parMap[parKey]));
		}

		mstch::map to_mustache_dict_recursive (
			const EntryNode& parNode,
			std::string_view parSrc,
			XPathRunner& parRunner
		);

		void store_entry_subtree (
			const std::vector<StructItem>& parXPaths,
			EntryNode& parCurrList
		) {
			for (auto& itm : parXPaths) {
				StructItemExtractor extractor(parCurrList);
				boost::apply_visitor(extractor, itm);
			}
		}

		class DictBuilder : public boost::static_visitor<> {
		public:
			explicit DictBuilder() :
				m_current_mustache_name(nullptr),
				m_current_mustache(nullptr)
			{
			}

			void operator() (const std::vector<ScrapNode>& parVal) {
				for (auto& val : parVal) {
					boost::apply_visitor(*this, val);
				}
			}

			void operator() (const FromBlock& parVal) {
#if defined(APPLY_VERBOSE)
				std::cout << parVal << '\n';
#endif
				m_global_entries.emplace_back(std::make_pair(
					&parVal.source,
					EntryNode("")
				));

				EntryNode& curr_node = m_global_entries.back().second;
				store_entry_subtree(parVal.xpaths, curr_node);
			}

			void operator() (const ApplyBlock& parVal) {
#if defined(APPLY_VERBOSE)
				std::cout << parVal << '\n';
#endif
				assert(not parVal.source.value.empty());
				m_apply_entries.emplace_back(&parVal.source, parVal.mustache_model);
				store_entry_subtree(parVal.xpaths, m_apply_entries.back().content);
			}

			void operator() (const MustacheBlock& parVal) {
#if defined(APPLY_VERBOSE)
				std::cout << "Mustache block \"" << parVal.name << "\"\n";
#endif
				const auto curr_name = m_current_mustache_name;

				if (not curr_name or *curr_name != parVal.name) {
					m_mustaches[parVal.name] = MustacheEntry();
					auto it_found = m_mustaches.find(parVal.name);
					m_current_mustache_name = &it_found->first;
					m_current_mustache = &it_found->second;
				}

				m_current_mustache->text = parVal.content;
			}

			const EntryNodeList& global_entries() const { return m_global_entries; }
			const MustacheEntryMap& mustache_entries() const { return m_mustaches; }
			const std::vector<ApplyEntry>& apply_entries() const { return m_apply_entries; }

		private:
			EntryNodeList m_global_entries;
			std::vector<ApplyEntry> m_apply_entries;
			MustacheEntryMap m_mustaches;
			const std::string* m_current_mustache_name;
			MustacheEntry* m_current_mustache;
		};

		struct ItemCountingVisitor : public boost::static_visitor<std::size_t> {
			template <typename T>
			std::size_t operator()(const T&) const { return 1; }
			std::size_t operator()(const mstch::array& parItem) const { return parItem.size(); }
		};

		class ArraysToStructArrayVisitor : public boost::static_visitor<> {
		public:
			explicit ArraysToStructArrayVisitor (std::size_t parExpectedSize) :
				m_expected_size(parExpectedSize)
			{
				m_array.resize(m_expected_size, mstch::map());
			}

			void operator()(const std::string& parName, const mstch::array& parItem) {
				for (std::size_t z = 0; z < parItem.size(); ++z) {
					auto& curr_map = boost::get<mstch::map>(m_array[z]);
					curr_map[parName] = parItem[z];
				}
			}

			template <typename T>
			void operator()(const std::string& parName, const T& parItem) {
				auto& curr_map = boost::get<mstch::map>(m_array[0]);
				curr_map[parName] = parItem;
			}

			mstch::node steal_struct() {
				if (1 == m_expected_size)
					return mstch::node(std::move(m_array[0]));
				else
					return mstch::node(std::move(m_array));
			}

		private:
			mstch::array m_array;
			const std::size_t m_expected_size;
		};

		const std::vector<std::string>& query_xpath_by_name (
			const EntryNodeList& parNodes,
			const std::string_view& parName,
			XPathRunner& parRunner
		) {
			for (auto& curr_node : parNodes) {
				assert(curr_node.first);
				const SourceInfo& source = *curr_node.first;
				const EntryNode& entry = curr_node.second;
				assert(entry.name.empty());

				auto it_found = std::find_if(
					entry.xpaths.begin(),
					entry.xpaths.end(),
					[&parName](const auto& xpath_elem) {
						return xpath_elem->name == parName;
					}
				);

				if (it_found != entry.xpaths.end()) {
					const XPathElement* const val = *it_found;
					assert(val);
					return parRunner.query(source.value, val->xpath);
				}
			}

			static const std::vector<std::string> empty_retval;
			std::cout << "query_xpath_by_name(parNodes, \"" << parName <<
				"\", parRunner) -> nothing found" << std::endl;
			assert(false); //throw?
			return empty_retval;
		}

		std::size_t largest_array_size_in (const mstch::map& parMap) {
			typedef ItemCountingVisitor ITC;
			using boost::apply_visitor;

			if (parMap.empty())
				return 0;

			return apply_visitor(
				ITC(),
				std::max_element(parMap.begin(), parMap.end(), [](const auto& a, const auto& b) {
					return apply_visitor(ITC(), a.second) < apply_visitor(ITC(), b.second);
				})->second
			);
		}

		void fill_with_xpaths (
			mstch::map& parOut,
			std::vector<FixLengthCommand>& parFixCommands,
			const EntryNode& parNode,
			std::string_view parSrc,
			XPathRunner& parRunner
		) {
			using std::placeholders::_1;

			for (const XPathElement* xpath : parNode.xpaths) {
				assert(xpath);
				std::cout << "Running query for \"" << xpath->name << "\"\n";
				auto results = parRunner.query(parSrc, xpath->xpath);
				if (results.size() == 1) {
					parOut[xpath->name] = results.front();
				}
				else if (results.size() > 1) {
					mstch::array values;
					values.reserve(results.size());
					std::copy(results.begin(), results.end(), std::back_inserter(values));
					parOut[xpath->name] = std::move(values);
				}
				else if (xpath->def_val) {
					parOut[xpath->name] = *xpath->def_val;
				}
				else {
					parOut[xpath->name] = std::string();
				}

				if (xpath->def_val)
					parFixCommands.push_back(std::bind(&fill_with_defaults, std::ref(parOut), _1, std::cref(xpath->name), std::cref(*xpath->def_val)));
				else
					parFixCommands.push_back(std::bind(&fill_with_last_item_clones, std::ref(parOut), _1, std::cref(xpath->name)));
			}
		}

		void fill_with_structs (
			mstch::map& parOut,
			const EntryNode& parNode,
			std::string_view parSrc,
			XPathRunner& parRunner
		) {
			for (auto& curr_struct : parNode.structs) {
				assert(not curr_struct.name.empty());

				auto new_struct = to_mustache_dict_recursive(curr_struct, parSrc, parRunner);
				const std::size_t extracted_struct_size = largest_array_size_in(new_struct);

				ArraysToStructArrayVisitor fix_visitor(extracted_struct_size);
				for (auto&& itm : new_struct) {
					auto visitor = [&fix_visitor,&name=itm.first](const auto& var) { fix_visitor(name, var); };
					boost::apply_visitor(visitor, itm.second);
				}

				parOut[std::string(curr_struct.name)] = fix_visitor.steal_struct();
			}
		}

		mstch::map to_mustache_dict_recursive (
			const EntryNode& parNode,
			std::string_view parSrc,
			XPathRunner& parRunner
		) {
			mstch::map retval;
			std::vector<FixLengthCommand> fix_commands;

			fill_with_xpaths(retval, fix_commands, parNode, parSrc, parRunner);
			fill_with_structs(retval, parNode, parSrc, parRunner);

			const std::size_t largest = largest_array_size_in(retval);
			for (const auto& command : fix_commands) {
				command(largest);
			}

			return retval;
		}

		mstch::map to_mustache_map (const EntryNodeList& parNodes, XPathRunner& parRunner) {
			mstch::map retval;
			for (auto& entry : parNodes) {
				assert(entry.second.name.empty());
				std::cout << "Analyzing entry " << *entry.first << '\n';

				assert(entry.first);
				std::string_view src_url;

				switch (entry.first->type) {
				case SourceInfo::URL:
					src_url = entry.first->value;
					break;
				case SourceInfo::Token:
				default:
					assert(false); //not reached
				}

				mstch::map curr_entry_map = to_mustache_dict_recursive(entry.second, src_url, parRunner);
				curr_entry_map.merge(std::move(retval));
				retval.swap(curr_entry_map);
			}

			return retval;
		}

		void exec_apply_block (
			const SourceInfo& parSourceInfo,
			const EntryNode& parEntryNode,
			const MustacheEntry& parMustache,
			XPathRunner& parXPathRunner
		) {
			EntryNodeList entry_node {std::make_pair(&parSourceInfo, parEntryNode)};
			mstch::map entry_ctx = to_mustache_map(entry_node, parXPathRunner);
			for (auto& ctx : parMustache.context) {
				entry_ctx[ctx.first] = ctx.second;
			}

			std::cout << "context size: " << entry_ctx.size() << '\n';
			for (auto& ctx_itm : entry_ctx) {
				std::cout << '\t' << ctx_itm.first << '\n';
			}

			std::cout << mstch::render(parMustache.text, entry_ctx) << std::endl;
		}
	} //unnamed namespace

	std::vector<std::string> apply (
		const ScrapNode& node,
		HtmlPoolBaseSP html_pool,
		XPathPtr xpath,
		std::string&& parDefNamespace
	) {
		using std::placeholders::_1;

		DictBuilder dict_builder;
		boost::apply_visitor(dict_builder, node);

		std::vector<std::string> retval;
		const EntryNodeList& global_entries = dict_builder.global_entries();
		const MustacheEntryMap& mustaches = dict_builder.mustache_entries();
		const std::vector<ApplyEntry> apply_entries = dict_builder.apply_entries();
		retval.reserve(apply_entries.size());

		std::cout << "-------------- visiting done ----------------\n";
		XPathRunner xpath_runner(html_pool, xpath, std::move(parDefNamespace));

		for (auto& apply_entry : apply_entries) {
			std::string name(apply_entry.mustache_name);
			const auto& mustache = mustaches.at(name);
			if (SourceInfo::Token == apply_entry.apply_to->type) {
				std::vector<std::string> sources =
					query_xpath_by_name(global_entries, apply_entry.apply_to->value, xpath_runner);

				for (auto& source : sources) {
					SourceInfo new_source;
					new_source.value = source;
					new_source.type = SourceInfo::URL;

					EntryNode new_node(apply_entry.content.name);
					new_node.structs = apply_entry.content.structs;
					new_node.xpaths = apply_entry.content.xpaths;

					exec_apply_block(new_source, new_node, mustache, xpath_runner);
				}
			}
			else {
				assert(apply_entry.apply_to);
				exec_apply_block(*apply_entry.apply_to, apply_entry.content, mustache, xpath_runner);
			}
		}

		return retval;
	}
}} //namespace duck::sl
