/*  Copyright (C) 2015 Michele Santullo
 *
 *  This file is part of DuckScraper.
 *
 *  DuckScraper is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  DuckScraper is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with DuckScraper.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef id1A180A0568E84FD88D57FAB82C69600E
#define id1A180A0568E84FD88D57FAB82C69600E

#include "SaltedIndex.hpp"
#include <map>
#include <cstdint>
#include <cassert>
#include <type_traits>
#include <vector>
#include <algorithm>

namespace duckutil {
	namespace Implem {
		template <typename Res, typename Name, typename IDT>
		class ResourceResNameWrapper {
		public:
			typedef IDT IDType;

			ResourceResNameWrapper ( const Name* parName, Res* parRes, IDType parID );
			ResourceResNameWrapper ( const ResourceResNameWrapper& ) = delete;
			~ResourceResNameWrapper ( void ) { assert(0 == m_refcount); }

			ResourceResNameWrapper& operator= (const ResourceResNameWrapper&) = delete;

			Res& GetResource ( void ) { return *m_resource; }
			const Res& GetResource ( void ) const { return *m_resource; }
			const Name& GetName ( void ) const { return *m_name; }
			void Retain ( void ) { ++m_refcount; }
			bool Release ( void ) { assert(m_refcount > 0); --m_refcount; return (0 == m_refcount); }
			uint32_t GetRefCount ( void ) const { return m_refcount; }
			uint32_t GetResourceID ( void ) const { return m_resId; }
			bool IsEmpty ( void ) const { return NULL == m_resource; }
			void Reset ( void ) { m_resource = NULL; m_name = NULL; m_refcount = m_resId = 0; }
			void DropRefCount ( void ) { m_refcount = 0; }

			bool operator== ( const ResourceResNameWrapper& parOther ) const { return (GetName() == parOther.GetName()); }
			bool operator!= ( const ResourceResNameWrapper& parOther ) const { return not (GetName() == parOther.GetName()); }
			bool operator<  ( const ResourceResNameWrapper& parOther ) const { return (GetName() < parOther.GetName()); }
			bool operator>  ( const ResourceResNameWrapper& parOther ) const { return (parOther.GetName() < GetName()); }
			bool operator>= ( const ResourceResNameWrapper& parOther ) const { return not (GetName() < parOther.GetName()); }
			bool operator<= ( const ResourceResNameWrapper& parOther ) const { return not (parOther.GetName() < GetName()); }

		private:
			Res* m_resource;
			const Name* m_name;
			uint16_t m_refcount;
			uint16_t m_resId;
		};
	} //namespace Implem

	template <typename Res, typename Name, typename Object=Name>
	class ResourcePool {
	public:
		typedef uint32_t IDType;
		typedef Name ResourceNameType;
	private:
		typedef Implem::ResourceResNameWrapper<Res, Name, IDType> ResourceWrapperType;
		typedef std::map<Name, ResourceWrapperType*> ResourceMapType;
		typedef std::vector<ResourceWrapperType*> ResourceVectorType;
	protected:
		typedef typename std::conditional<std::is_fundamental<ResourceNameType>::value, ResourceNameType, const ResourceNameType&>::type ResourceNameParamType;
		typedef typename std::conditional<std::is_fundamental<Object>::value, Object, const Object&>::type ResourceObjectParameterType;
	public:
		typedef Res ResourceType;
		typedef Object ResourceObjectType;

		ResourcePool ( void ) = default;
		ResourcePool ( const ResourcePool& ) = delete;
		virtual ~ResourcePool ( void ) = default;

		ResourcePool& operator= (const ResourcePool&) = delete;

		ResourceType* GetByName ( ResourceNameParamType parName );
		const ResourceType* GetByName ( ResourceNameParamType parName ) const;
		IDType GetOrAdd ( ResourceObjectParameterType parObjectName );
		ResourceType* GetByID ( IDType parID );
		const ResourceType* GetByID ( IDType parID ) const;
		bool IsEmpty ( void ) const;

		IDType AddResource ( ResourceObjectParameterType parRes );
		void ReleaseResource ( IDType parRes );
		void ReleaseResourceByName ( ResourceNameParamType parName );

		void Dispose ( void ) noexcept;

	protected:
		void Dispose_IgnoreReferenceCount ( void );

		virtual ResourceType* OnResourceLoad ( ResourceObjectParameterType parRes ) = 0;
		virtual void OnResourceDestroy ( ResourceNameParamType parName, ResourceType* parRes ) noexcept = 0;
		virtual ResourceNameType GetResourceNameFromResourceObject ( ResourceObjectParameterType parRes ) = 0;

	private:
		bool ReleaseResource ( typename ResourceMapType::iterator& parMapRes, typename ResourceVectorType::iterator& parVecRes );

		ResourceMapType m_mapContainer; //For accesses by name
		ResourceVectorType m_linearContainer; //For accesses by index
	};
} //namespace duckutil

#include "ResourcePool.inl"

#endif
