/*  Copyright (C) 2015 Michele Santullo
 *
 *  This file is part of DuckScraper.
 *
 *  DuckScraper is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  DuckScraper is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with DuckScraper.  If not, see <http://www.gnu.org/licenses/>.
 */

namespace duckutil {
	namespace Implem {
		///---------------------------------------------------------------------
		///---------------------------------------------------------------------
		template <typename Res, typename Name, typename IDT>
		ResourceResNameWrapper<Res, Name, IDT>::ResourceResNameWrapper (const Name* parName, Res* parRes, IDType parID) {
			assert(nullptr != parRes);
			assert(nullptr != parName);

			m_resource = parRes;
			m_name = parName;
			m_refcount = 0;
			m_resId = static_cast<uint16_t>(parID);
		}

		///---------------------------------------------------------------------
		///---------------------------------------------------------------------
		template <typename V>
		inline void TrimTrailingNulls (V& parVector) {
			const std::size_t nullsCount = std::find_if(parVector.rbegin(), parVector.rend(), std::bind1st(std::not_equal_to<typename V::value_type>(), nullptr)) - parVector.rbegin();
			assert(nullsCount <= parVector.size());
			if (nullsCount) {
				assert(nullptr == parVector.back());
				parVector.resize(parVector.size() - nullsCount);
			}
		}
	} //namespace Implem

	///-------------------------------------------------------------------------
	///-------------------------------------------------------------------------
	template <typename Res, typename Name, typename Object>
	typename ResourcePool<Res, Name, Object>::IDType ResourcePool<Res, Name, Object>::GetOrAdd (ResourceObjectParameterType parObjectName) {
		const ResourceNameType name = GetResourceNameFromResourceObject(parObjectName);
		typename ResourceMapType::const_iterator itFind = m_mapContainer.find(name);
		IDType retVal;
		if (m_mapContainer.end() == itFind) {
			retVal = AddResource(parObjectName);
		}
		else {
			typename ResourceVectorType::const_iterator itVecFind = std::find(m_linearContainer.begin(), m_linearContainer.end(), itFind->second);
			assert(m_linearContainer.end() != itVecFind);
			retVal = static_cast<IDType>(itVecFind - m_linearContainer.begin() + 1);
		}
		return retVal;
	}

	///-------------------------------------------------------------------------
	///-------------------------------------------------------------------------
	template <typename Res, typename Name, typename Object>
	typename ResourcePool<Res, Name, Object>::ResourceType* ResourcePool<Res, Name, Object>::GetByName (ResourceNameParamType parName) {
		typename ResourceMapType::iterator itFind = m_mapContainer.find(parName);
		if (m_mapContainer.end() == itFind)
			return nullptr;
		else
			return &itFind->second->GetResource();
	}

	///-------------------------------------------------------------------------
	///-------------------------------------------------------------------------
	template <typename Res, typename Name, typename Object>
	const typename ResourcePool<Res, Name, Object>::ResourceType* ResourcePool<Res, Name, Object>::GetByName (ResourceNameParamType parName) const {
		typename ResourceMapType::const_iterator itFind = m_mapContainer.find(parName);
		if (m_mapContainer.end() == itFind)
			return nullptr;
		else
			return &itFind->second->GetResource();
	}

	///-------------------------------------------------------------------------
	///-------------------------------------------------------------------------
	template <typename Res, typename Name, typename Object>
	typename ResourcePool<Res, Name, Object>::ResourceType* ResourcePool<Res, Name, Object>::GetByID (IDType parID) {
		assert(parID > 0);
		if (0 == parID)
			return nullptr;

		const auto index = static_cast<std::size_t>(parID - 1);
		if (index < m_linearContainer.size()) {
			ResourceWrapperType* res = m_linearContainer[index];
			return &res->GetResource();
		}
		else {
			return nullptr;
		}
	}

	///-------------------------------------------------------------------------
	///-------------------------------------------------------------------------
	template <typename Res, typename Name, typename Object>
	const typename ResourcePool<Res, Name, Object>::ResourceType* ResourcePool<Res, Name, Object>::GetByID (IDType parID) const {
		assert(parID > 0);
		if (0 == parID)
			return nullptr;

		const auto index = static_cast<std::size_t>(parID - 1);
		if (index < m_linearContainer.size()) {
			ResourceWrapperType* res = m_linearContainer[index];
			return &res->GetResource();
		}
		else {
			return nullptr;
		}
	}

	///-------------------------------------------------------------------------
	///-------------------------------------------------------------------------
	template <typename Res, typename Name, typename Object>
	typename ResourcePool<Res, Name, Object>::IDType ResourcePool<Res, Name, Object>::AddResource (ResourceObjectParameterType parRes) {
		const ResourceNameType name = GetResourceNameFromResourceObject(parRes);
		typename ResourceMapType::iterator itPreExisting = m_mapContainer.find(name);
		if (m_mapContainer.end() != itPreExisting) {
// 				if (itPreExisting->IsEmpty()) {
// 					OnResourceReload(name);
			assert(nullptr != itPreExisting->second);
			assert(not itPreExisting->second->IsEmpty());
			itPreExisting->second->Retain();
			return itPreExisting->second->GetResourceID();
		}
		else {
			assert(m_mapContainer.end() == m_mapContainer.find(name));
			ResourceType* const newRes = OnResourceLoad(parRes);
			if (newRes) {
				std::pair<typename ResourceMapType::iterator, bool> newIt = m_mapContainer.insert(std::pair<ResourceNameType, ResourceWrapperType*>(name, nullptr));

				IDType newID = static_cast<IDType>(m_linearContainer.size() + 1);
				ResourceWrapperType* const newWrapper = new ResourceWrapperType(&newIt.first->first, newRes, newID);
				assert(nullptr != newWrapper);
				m_linearContainer.push_back(newWrapper);
				newIt.first->second = newWrapper;
				newWrapper->Retain();
				return newID;
			}
		}
		return 0;
	}

	///-------------------------------------------------------------------------
	///-------------------------------------------------------------------------
	template <typename Res, typename Name, typename Object>
	void ResourcePool<Res, Name, Object>::ReleaseResource (IDType parRes) {
		assert(parRes > 0);
		assert(static_cast<std::size_t>(parRes) <= m_linearContainer.size());

		typename ResourceVectorType::iterator rele = m_linearContainer.begin() + (parRes - 1);
		assert(nullptr != *rele);
		assert(rele - m_linearContainer.begin() == static_cast<int>(parRes - 1));

		typename ResourceMapType::iterator relemap = m_mapContainer.find((*rele)->GetName());
		assert(m_mapContainer.end() != relemap);

		if (ReleaseResource(relemap, rele)) {
			delete relemap->second;
			m_mapContainer.erase(relemap);
			Implem::TrimTrailingNulls(m_linearContainer);
		}
	}

	///-------------------------------------------------------------------------
	///-------------------------------------------------------------------------
	template <typename Res, typename Name, typename Object>
	void ResourcePool<Res, Name, Object>::ReleaseResourceByName (ResourceNameParamType parName) {
		typename ResourceMapType::iterator rele = m_mapContainer.find(parName);
		assert(m_mapContainer.end() != rele);

		const IDType resId = rele->second->GetResourceID();
		assert(static_cast<std::size_t>(resId) <= m_linearContainer.size());
		assert(resId > 0);

		if (ReleaseResource(rele, m_linearContainer.begin() + (resId - 1))) {
			delete rele->second;
			m_mapContainer.erase(rele);
			Implem::TrimTrailingNulls(m_linearContainer);
		}
	}

	///-------------------------------------------------------------------------
	///-------------------------------------------------------------------------
	template <typename Res, typename Name, typename Object>
	bool ResourcePool<Res, Name, Object>::ReleaseResource (typename ResourceMapType::iterator& parMapRes, typename ResourceVectorType::iterator& parVecRes) {
		assert(parMapRes->second == *parVecRes);
		assert(nullptr != *parVecRes);
		ResourceWrapperType& currRes = **parVecRes;

		assert(not currRes.IsEmpty());
		if (not currRes.IsEmpty()) {
			if (currRes.Release()) {
				this->OnResourceDestroy(currRes.GetName(), &currRes.GetResource());
				currRes.Reset();
				*parVecRes = nullptr;
				return true;
			}
		}
		return false;
	}

	///-------------------------------------------------------------------------
	///-------------------------------------------------------------------------
	template <typename Res, typename Name, typename Object>
	bool ResourcePool<Res, Name, Object>::IsEmpty() const {
		return m_mapContainer.empty() and m_linearContainer.empty();
	}

	///-------------------------------------------------------------------------
	///-------------------------------------------------------------------------
	template <typename Res, typename Name, typename Object>
	void ResourcePool<Res, Name, Object>::Dispose() noexcept {
		for (auto& currItem : m_linearContainer) {
			if (nullptr != currItem)
				this->OnResourceDestroy(currItem->GetName(), &currItem->GetResource());
			delete currItem;
		}
		m_linearContainer.clear();
		m_mapContainer.clear();
	}

	///-------------------------------------------------------------------------
	///-------------------------------------------------------------------------
	template <typename Res, typename Name, typename Object>
	void ResourcePool<Res, Name, Object>::Dispose_IgnoreReferenceCount() {
		for (auto& currItem : m_linearContainer) {
			currItem->DropRefCount();
		}
		Dispose();
	}
} //namespace duckcore
