/*  Copyright (C) 2017-2020 Michele Santullo
 *
 *  This file is part of DuckScraper.
 *
 *  DuckScraper is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  DuckScraper is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with DuckScraper.  If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include <boost/spirit/include/qi.hpp>
#include <boost/spirit/include/phoenix_operator.hpp>

namespace kamokan {
	template <typename Iterator>
	struct IniCommentSkipper : boost::spirit::qi::grammar<Iterator> {
		IniCommentSkipper() :
			IniCommentSkipper::base_type(skipping),
			first_char(true)
		{
			namespace px = boost::phoenix;
			using boost::spirit::qi::blank;
			using boost::spirit::qi::lit;
			using boost::spirit::qi::eol;
			using boost::spirit::qi::char_;
			using boost::spirit::qi::eps;

			skipping = comment | blank;
			comment = (eps(px::cref(first_char) == true) | eol) >>
				*blank >> lit("#")[px::ref(first_char) = false] >>
				*(!eol >> char_);
		}

		boost::spirit::qi::rule<Iterator> skipping;
		boost::spirit::qi::rule<Iterator> comment;
		bool first_char;
	};
} //namespace kamokan

namespace duck::sl {
	namespace qi = ::boost::spirit::qi;

	template <typename I, typename Skipper>
	class ScrapGrammar : public qi::grammar<I, std::vector<ScrapNode>(), Skipper> {
	public:
		ScrapGrammar() : ScrapGrammar::base_type(start) {
			using qi::char_;
			using qi::lexeme;
			using qi::alpha;
			using qi::alnum;
			using qi::graph;
			using qi::attr;
			using qi::eol;
			using qi::eoi;
			using qi::lit;
			using qi::string;
			using qi::as_string;
			using qi::no_skip;

			start = *eol >> (from_block | apply_block | mustache_block) % +eol >> *eol >> eoi;
			from_block = lit("from") >> source_info >> +eol >> assignment_list >> +eol >> "end";
			source_info = ((url | string("-")) >> attr(SourceInfo::URL)) | (mustache_like_token >> attr(SourceInfo::Token));
			url = -(+alpha >> string("://")) >> alpha >> *graph;
			mustache_like_token = "{{" >> identifier >> "}}";
			quoted_string %= lexeme['"' >> *(char_ - '"') >> '"'];
			xpath_assignment %= identifier >>
				-(lit("default") >> '(' >> quoted_string >> ')') >> "=" >>
				as_string[lexeme[+(graph | char_(" \t"))]];
			identifier %= lexeme[(alpha | char_('_')) >> *(-char_('.') >> +(alnum | char_('_')))];

			apply_block = lit("apply") >> mustache_like_token >> "to" >> source_info >> +eol >>
				assignment_list >> +eol >> "end";
			struct_block = "struct" >> identifier >> +eol >> assignment_list >> +eol >> "end";

			mustache_block %= as_string[lit("==") >> identifier] >> eol >>
				as_string[no_skip[+(!lit("==end") >> char_)]] >> "==end";

			assignment_list = (xpath_assignment | struct_block) % +eol;
		}

	private:
		template <typename F>
		using RuleType = qi::rule<I, F, Skipper>;

		RuleType<std::vector<ScrapNode>()> start;
		RuleType<FromBlock()> from_block;
		RuleType<std::string()> url;
		RuleType<std::string()> mustache_like_token;
		RuleType<std::string()> quoted_string;
		RuleType<XPathElement()> xpath_assignment;
		RuleType<std::string()> identifier;
		RuleType<SourceInfo()> source_info;
		RuleType<ApplyBlock()> apply_block;
		RuleType<StructBlock()> struct_block;
		RuleType<MustacheBlock()> mustache_block;
		RuleType<std::vector<StructItem>()> assignment_list;
	};
} //namespace duck::sl
