/*  Copyright (C) 2015 Michele Santullo
 *
 *  This file is part of DuckScraper.
 *
 *  DuckScraper is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  DuckScraper is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with DuckScraper.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "xpath.hpp"
#include <xercesc/framework/MemBufInputSource.hpp>
#include <xercesc/util/XMLString.hpp>
#include <xqilla/exceptions/XQException.hpp>
#include <xqilla/exceptions/XMLParseException.hpp>
#include <xqilla/context/ContextHelpers.hpp>
#include <sstream>
#include <stdexcept>
#include <algorithm>

namespace duck {
	namespace {
		typedef std::pair<int, int> LineColType;

		LineColType line_col_from_offset (ptrdiff_t parOffset, const std::string& parData) {
			size_t index = 0;
			int line = 1;
			int chara = 1;
			while (parOffset and index < parData.size()) {
				if (parData[index] == '\n') {
					chara = 1;
					++line;
				}
				else {
					++chara;
				}
				++index;
				--parOffset;
			}
			return std::make_pair(line, chara);
		}
	} //unnamed namespace

	XPath::XPath() = default;

	XPath::~XPath() = default;

	auto XPath::run_query (const std::string& parXML, const std::vector<std::string>& parQueries, const std::string& parDefNamespace) -> BatchResults {
		XQilla& xqilla = m_xqilla;
		XercesConfiguration xconfig;
		xercesc::MemBufInputSource input_buf(reinterpret_cast<const XMLByte*>(parXML.c_str()), parXML.size(), "n/a", false);
		BatchResults retval;
		try {
			AutoDelete<DynamicContext> context(xqilla.createContext(XQilla::XQUERY3, &xconfig));
			xconfig.populateStaticContext(context);
			Node::Ptr ptr = context->parseDocument(input_buf);
			context->setContextItem(ptr);
			//see http://xqilla.sourceforge.net/docs/simple-api/classStaticContext.html#adc869a84712459fa49db67fe837c9b01
			AutoDeleteArray<XMLCh> ns_wide = xercesc::XMLString::transcode(parDefNamespace.c_str());
			context->setDefaultElementAndTypeNS(ns_wide);

			for (const auto& xpath : parQueries) {
				AutoDelete<XQQuery> query(nullptr);
				{
					AutoContextInfoReset resetter(context);
					AutoDeleteArray<XMLCh> xpath_wide = xercesc::XMLString::transcode(xpath.c_str());
					query.set(xqilla.parse(xpath_wide, context, nullptr, XQilla::NO_ADOPT_CONTEXT));
				}

				Result result = query->execute(context);
				Item::Ptr item;
				std::vector<std::pair<std::string, std::string>> new_lst;
				while(nullptr != (item = result->next(context))) {
					new_lst.push_back(std::make_pair(std::string(), UTF8(item->asString(context))));
				}

				retval.push_back(std::move(new_lst));
			}
		}
		catch (const XMLParseException& err) {
			throw ParseError(err.getXQueryLine(), err.getXQueryColumn(), xercesc::XMLString::transcode(err.getError()));
		}
		catch (const XQException& err) {
			throw std::runtime_error(xercesc::XMLString::transcode(err.getError()));
		}

		return retval;
	}

	std::vector<std::string> XPath::run_query (const std::string& parXML, const std::string& parQuery, const std::string& parDefNamespace) {
		auto query_res = run_query(parXML, std::vector<std::string>{parQuery}, parDefNamespace);
		if (query_res.empty() or query_res.front().empty()) {
			return std::vector<std::string>();
		}
		else {
			std::vector<std::string> retval;
			const std::vector<std::pair<std::string, std::string>>& src = query_res.front();
			retval.reserve(src.size());
			std::transform(src.begin(), src.end(), std::back_inserter(retval), [](const auto& pair) { return pair.second; });
			return retval;
		}
	}

	ParseError::ParseError (int parLine, int parColumn, std::string parMessage) {
		std::ostringstream oss;
		oss << "Error parsing the source XML at line " <<
			parLine << " col " << parColumn << ":\n" <<
			parMessage << std::endl;
		auto msg = oss.str();
		m_msg.resize(msg.size() + 1);
		std::copy(msg.begin(), msg.end(), m_msg.begin());
		m_msg[m_msg.size() - 1] = '\0';
	}

	const char* ParseError::what() const noexcept {
		return m_msg.data();
	}
} //namespace duck
